from __future__ import annotations

import concurrent.futures
import random
from abc import abstractmethod, ABC
from enum import Enum
from multiprocessing import cpu_count
from typing import List, Union, Tuple

from ttlinks.ipservice.ip_addr_type_classifiers import IPv4AddrClassifierPublicHandler
from ttlinks.ipservice.ip_address import IPv4Addr, IPv4NetMask, IPAddr, IPv6NetMask, IPv6Addr
from ttlinks.ipservice.ip_configs import InterfaceIPConfig, IPv4HostConfig, IPv4SubnetConfig, IPv4WildCardConfig, IPv6HostConfig, IPv6SubnetConfig, \
    IPv6WildCardConfig
from ttlinks.ipservice.ip_converters import DecimalIPv4ConverterHandler, DecimalIPv6ConverterHandler
from ttlinks.ipservice.ip_utils import IPv4AddrType


class IPv4TypeAddrBlocks(Enum):
    UNSPECIFIED = ['0.0.0.0/32']
    CURRENT_NETWORK = ['0.0.0.0/8']
    PUBLIC = [
            '0.0.0.0/8', '10.0.0.0/8', '100.64.0.0/10', '127.0.0.0/8', '169.254.0.0/16',
            '172.16.0.0/12', '192.0.0.0/24', '192.0.2.0/24', '192.88.99.0/24', '192.168.0.0/16',
            '198.18.0.0/15', '198.51.100.0/24', '203.0.113.0/24', '224.0.0.0/4', '233.252.0.0/24',
            '240.0.0.0/4', '255.255.255.255/32'
        ]  # Public IPs generated have to avoid the blocks in the list.
    PRIVATE = ['10.0.0.0/8', '172.16.0.0/12', '192.168.0.0/16']
    MULTICAST = ["224.0.0.0/4"]  # For 224.0.0.0/4 and others
    LINK_LOCAL = ["169.254.0.0/16"]  # For 169.254.0.0/16
    LOOPBACK = ["127.0.0.0/8"]  # For 127.0.0.0/8
    DOCUMENTATION = ['192.0.2.0/24', '198.51.100.0/24', '203.0.113.0/24', '233.252.0.0/24']  # For 192.0.2.0/24, 198.51.100.0/24, 203.0.113.0/24, and 233.252.0.0/24
    CARRIER_GRADE_NAT = ['100.64.0.0/10']  # For 100.64.0.0/10
    BENCHMARK_TESTING = ['198.18.0.0/15']  # For 198.18.0.0/15
    IPV6_TO_IPV4_RELAY = ['192.88.99.0/24']  # For 192.88.99.0/24 (formerly used)
    RESERVED = ['240.0.0.0/4']  # For 240.0.0.0/4 and 255.255.255.255/32
    LIMITED_BROADCAST = ['255.255.255.255/32']  # For 255.255.255.255/32, "limited broadcast" destination address
    DS_LITE = ['192.0.0.0/24']  # For 192.0.0.0/24

class IPv6TypeAddrBlocks(Enum):
    UNSPECIFIED = ['::/128']  # ::/128
    SRV6 = ['5f00::/16']  # 5f00::/16, IPv6 Segment Routing
    GLOBAL_UNICAST = ['2000::/3']
    UNIQUE_LOCAL = ['fc00::/7']  # fc00::/7
    MULTICAST = ['ff00::/8']  # ff00::/8
    LINK_LOCAL = ['fe80::/64']  # fe80::/64 from fe80::/10
    LOOPBACK = ['::1/128']  # ::1/128
    DOCUMENTATION = ['2001:db8::/32', '3fff::/20']  # 2001:db8::/32 and 3fff::/20
    IPV4_MAPPED = ['::ffff:0:0/96']  # ::ffff:0:0/96
    IPV4_TRANSLATED = ['::ffff:0:0:0/96']  # ::ffff:0:0:0/96
    IPV4_IPV6_TRANSLATION = ['64:ff9b::/96', '64:ff9b:1::/48']  # 64:ff9b::/96 and 64:ff9b:1::/48
    DISCARD_PREFIX = ['100::/64']  # 100::/64
    IP6_TO4 = ['2002::/16'] # 2002::/16, 6to4 addressing scheme
    TEREDO_TUNNELING = ['2001::/32']  # 2001::/32
    ORCHIDV2 = ['2001:20::/28']  # 2001:20::/28


class IPFactory(ABC):
    """
    Abstract Base Class (ABC) for generating various IP configurations, including host IPs, subnets, and wildcard configurations.
    This class serves as the blueprint for concrete implementations that can generate random or batch configurations for different address types.

    Attributes:
    _subnet_cache (dict): A cache to store pre-generated subnet configurations, reducing the need for redundant calculations.
    """
    def __init__(self):
        """
        Initializes the IPFactory object by setting up a cache to store subnet configurations.
        This cache helps improve performance by reusing subnet data rather than recalculating it.
        """
        self._subnet_cache = {}

    @abstractmethod
    def host(self, host: str) -> InterfaceIPConfig:
        """
        Abstract method to generate a host IP configuration.

        Parameters:
        host (str): The string representation of the host IP.

        Returns:
        InterfaceIPConfig: The configuration object representing the host IP.
        """
        pass

    @abstractmethod
    def subnet(self, subnet: str) -> InterfaceIPConfig:
        """
        Abstract method to generate a subnet configuration.

        Parameters:
        subnet (str): The string representation of the subnet.

        Returns:
        InterfaceIPConfig: The configuration object representing the subnet.
        """
        pass

    @abstractmethod
    def wildcard(self, wildcard: str) -> InterfaceIPConfig:
        """
        Abstract method to generate a wildcard configuration.

        Parameters:
        wildcard (str): The string representation of the wildcard IP.

        Returns:
        InterfaceIPConfig: The configuration object representing the wildcard IP.
        """
        pass

    @abstractmethod
    def batch_hosts(self, hosts: List[str], max_workers: int = 10, keep_dup = True) -> List[InterfaceIPConfig]:
        """
        Abstract method to generate a batch of host configurations.

        Parameters:
        hosts (List[str]): A list of host IP strings.
        max_workers (int): The maximum number of workers to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate host IPs in the batch.

        Returns:
        List[InterfaceIPConfig]: A list of configuration objects representing the host IPs.
        """
        pass

    @abstractmethod
    def batch_subnets(self, subnets: List[str], max_workers: int = 10, keep_dup = True) -> List[InterfaceIPConfig]:
        """
        Abstract method to generate a batch of subnet configurations.

        Parameters:
        subnets (List[str]): A list of subnet IP strings.
        max_workers (int): The maximum number of workers to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate subnet IPs in the batch.

        Returns:
        List[InterfaceIPConfig]: A list of configuration objects representing the subnets.
        """
        pass

    @abstractmethod
    def random_host(self, addr_type=None) -> InterfaceIPConfig:
        """
        Abstract method to generate a random host IP configuration.

        Parameters:
        addr_type (Any): The type of address block for which to generate the random host IP.

        Returns:
        InterfaceIPConfig: The configuration object representing the random host IP.
        """
        pass

    @abstractmethod
    def random_subnet(self, addr_type=None) -> InterfaceIPConfig:
        """
        Abstract method to generate a random subnet configuration.

        Parameters:
        addr_type (Any): The type of address block for which to generate the random subnet.

        Returns:
        InterfaceIPConfig: The configuration object representing the random subnet.
        """
        pass

    @abstractmethod
    def random_hosts_batch(self, addr_type=None, num_ips = 10) -> List[InterfaceIPConfig]:
        """
        Abstract method to generate a batch of random host IP configurations.

        Parameters:
        addr_type (Any): The type of address block for which to generate the random host IPs.
        num_ips (int): The number of random IPs to generate in the batch.

        Returns:
        List[InterfaceIPConfig]: A list of configuration objects representing the random host IPs.
        """
        pass

    @abstractmethod
    def random_subnets_batch(self, addr_type=None, num_ips = 10) -> List[InterfaceIPConfig]:
        """
        Abstract method to generate a batch of random subnet configurations.

        Parameters:
        addr_type (Any): The type of address block for which to generate the random subnets.
        num_ips (int): The number of random subnets to generate in the batch.

        Returns:
        List[InterfaceIPConfig]: A list of configuration objects representing the random subnets.
        """
        pass

    @abstractmethod
    def _random_subnet_for_type(self, addr_type) -> InterfaceIPConfig:
        """
        Abstract method to generate a random subnet configuration for a specific address type.

        Parameters:
        addr_type (Any): The type of address block for which to generate the random subnet.

        Returns:
        InterfaceIPConfig: The configuration object representing the random subnet for the given address type.
        """
        pass



class IPv4Factory(IPFactory):
    """
    Concrete implementation of IPFactory for handling IPv4 addresses.
    This class provides methods to generate and manage IPv4 host, subnet, and wildcard configurations, as well as random IP generation.
    """
    def host(self, host: str) -> IPv4HostConfig:
        """
        Generates an IPv4 host configuration.

        Parameters:
        host (str): The string representation of the host IP.

        Returns:
        IPv4HostConfig: The configuration object representing the host IP.
        """
        return IPv4HostConfig(host)

    def subnet(self, subnet: str) -> IPv4SubnetConfig:
        """
        Generates an IPv4 subnet configuration.

        Parameters:
        subnet (str): The string representation of the subnet.

        Returns:
        IPv4SubnetConfig: The configuration object representing the subnet.
        """
        return IPv4SubnetConfig(subnet)

    def wildcard(self, wildcard: str) -> IPv4WildCardConfig:
        """
        Generates an IPv4 wildcard configuration.

        Parameters:
        wildcard (str): The string representation of the wildcard IP.

        Returns:
        IPv4WildCardConfig: The configuration object representing the wildcard IP.
        """
        return IPv4WildCardConfig(wildcard)

    def batch_hosts(self, hosts: List[str], max_workers: int = 10, keep_dup = True) -> List[IPv4HostConfig]:
        """
        Generates a batch of IPv4 host configurations using multi-threading.

        Parameters:
        hosts (List[str]): A list of host IP strings.
        max_workers (int): The maximum number of threads to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate host IPs in the batch.

        Returns:
        List[IPv4HostConfig]: A list of configuration objects representing the host IPs.
        """
        if keep_dup is False:
            hosts = list(set(hosts))
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = list(executor.map(self.host, hosts))
        return results

    def batch_subnets(self, subnets: List[str], max_workers: int = 10, keep_dup = True) -> List[IPv4SubnetConfig]:
        """
        Generates a batch of IPv4 subnet configurations using multi-threading.

        Parameters:
        subnets (List[str]): A list of subnet IP strings.
        max_workers (int): The maximum number of threads to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate subnet IPs in the batch.

        Returns:
        List[IPv4SubnetConfig]: A list of configuration objects representing the subnets.
        """
        if keep_dup is False:
            subnets = list(set(subnets))
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = list(executor.map(self.subnet, subnets))
        return results

    def random_host(self, addr_type:IPv4TypeAddrBlocks=None) -> IPv4HostConfig:
        """
        Generates a random IPv4 host configuration based on the specified address type.

        Parameters:
        addr_type (IPv4TypeAddrBlocks): The address block type to use for generating the random host.

        Returns:
        IPv4HostConfig: A randomly generated host configuration.
        """
        subnet = self._random_subnet_for_type(addr_type)
        generated_host = IPv4HostConfigRandomizer(subnet).randomize()
        if addr_type == IPv4TypeAddrBlocks.PUBLIC and IPv4AddrClassifierPublicHandler().handle(generated_host.ip_addr) != IPv4AddrType.PUBLIC:
            return self.random_host(addr_type)
        return generated_host

    def random_subnet(self, addr_type:IPv4TypeAddrBlocks=None) -> IPv4SubnetConfig:
        """
        Generates a random IPv4 subnet configuration based on the specified address type.

        Parameters:
        addr_type (IPv4TypeAddrBlocks): The address block type to use for generating the random subnet.

        Returns:
        IPv4SubnetConfig: A randomly generated subnet configuration.
        """
        subnet = self._random_subnet_for_type(addr_type)
        generated_subnet = IPv4SubnetConfigRandomizer(subnet).randomize()
        if addr_type == IPv4TypeAddrBlocks.PUBLIC and IPv4AddrClassifierPublicHandler().handle(generated_subnet.ip_addr) != IPv4AddrType.PUBLIC:
            return self.random_subnet(addr_type)
        return generated_subnet

    def random_hosts_batch(self, addr_type:IPv4TypeAddrBlocks=None, num_ips = 10) -> List[IPv4HostConfig]:
        """
        Generates a batch of random IPv4 host configurations using multiprocessing.

        Parameters:
        addr_type (IPv4TypeAddrBlocks): The address block type to use for generating the random hosts.
        num_ips (int): The number of random IPs to generate in the batch.

        Returns:
        List[IPv4HostConfig]: A list of randomly generated host configurations.
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu_count()) as executor:
            futures = [executor.submit(self.random_host, addr_type) for _ in range(num_ips)]
            random_ips = [future.result() for future in concurrent.futures.as_completed(futures)]
        return random_ips

    def random_subnets_batch(self, addr_type:IPv4TypeAddrBlocks=None, num_ips = 10) -> List[IPv4SubnetConfig]:
        """
        Generates a batch of random IPv4 subnet configurations using multiprocessing.

        Parameters:
        addr_type (IPv4TypeAddrBlocks): The address block type to use for generating the random subnets.
        num_ips (int): The number of random subnets to generate in the batch.

        Returns:
        List[IPv4SubnetConfig]: A list of randomly generated subnet configurations.
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu_count()) as executor:
            futures = [executor.submit(self.random_subnet, addr_type) for _ in range(num_ips)]
            random_subnets = [future.result() for future in concurrent.futures.as_completed(futures)]
        return random_subnets

    def _random_subnet_for_type(self, addr_type:IPv4TypeAddrBlocks) -> IPv4SubnetConfig:
        """
        Retrieves or generates a random subnet configuration for the given address type, utilizing caching for performance.

        Parameters:
        addr_type (IPv4TypeAddrBlocks): The address block type to use for generating the subnet.

        Returns:
        IPv4SubnetConfig: A random subnet configuration.
        """
        if addr_type in self._subnet_cache:
            subnets = self._subnet_cache[addr_type]
        else:
            if addr_type == IPv4TypeAddrBlocks.PUBLIC:
                subnets = self.batch_subnets(['0.0.0.0/0'])
            elif addr_type is None:
                subnets = self.batch_subnets(['0.0.0.0/0'])
            else:
                subnets = self.batch_subnets(addr_type.value)
            self._subnet_cache[addr_type] = subnets
        return random.choice(subnets)

class IPv6Factory(IPFactory):
    """
    Concrete implementation of IPFactory for handling IPv6 addresses.
    This class provides methods to generate and manage IPv6 host, subnet, and wildcard configurations, as well as random IP generation.
    """
    def host(self, host: str) -> IPv6HostConfig:
        """
        Generates an IPv6 host configuration.

        Parameters:
        host (str): The string representation of the host IPv6 address.

        Returns:
        IPv6HostConfig: The configuration object representing the host IPv6 address.
        """
        return IPv6HostConfig(host)

    def subnet(self, subnet: str) -> IPv6SubnetConfig:
        """
        Generates an IPv6 subnet configuration.

        Parameters:
        subnet (str): The string representation of the IPv6 subnet.

        Returns:
        IPv6SubnetConfig: The configuration object representing the IPv6 subnet.
        """
        return IPv6SubnetConfig(subnet)

    def wildcard(self, wildcard: str) -> IPv6WildCardConfig:
        """
        Generates an IPv6 wildcard configuration.

        Parameters:
        wildcard (str): The string representation of the wildcard IPv6 address.

        Returns:
        IPv6WildCardConfig: The configuration object representing the wildcard IPv6 address.
        """
        return IPv6WildCardConfig(wildcard)

    def batch_hosts(self, hosts: List[str], max_workers: int = 10, keep_dup = True) -> List[IPv6HostConfig]:
        """
        Generates a batch of IPv6 host configurations using multi-threading.

        Parameters:
        hosts (List[str]): A list of IPv6 host strings.
        max_workers (int): The maximum number of threads to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate IPv6 host addresses in the batch.

        Returns:
        List[IPv6HostConfig]: A list of configuration objects representing the host IPv6 addresses.
        """
        if keep_dup is False:
            hosts = list(set(hosts))
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = list(executor.map(self.host, hosts))
        return results

    def batch_subnets(self, subnets: List[str], max_workers: int = 10, keep_dup = True) -> List[IPv6SubnetConfig]:
        """
        Generates a batch of IPv6 subnet configurations using multi-threading.

        Parameters:
        subnets (List[str]): A list of IPv6 subnet strings.
        max_workers (int): The maximum number of threads to use for parallel processing.
        keep_dup (bool): Whether to keep duplicate IPv6 subnet addresses in the batch.

        Returns:
        List[IPv6SubnetConfig]: A list of configuration objects representing the IPv6 subnets.
        """
        if keep_dup is False:
            subnets = list(set(subnets))
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = list(executor.map(self.subnet, subnets))
        return results

    def random_host(self, addr_type=None) -> IPv6HostConfig:
        """
        Generates a random IPv6 host configuration based on the specified address type.

        Parameters:
        addr_type (Any): The address block type to use for generating the random IPv6 host.

        Returns:
        IPv6HostConfig: A randomly generated IPv6 host configuration.
        """
        subnet = self._random_subnet_for_type(addr_type)
        generated_host = IPv6HostConfigRandomizer(subnet).randomize()
        return generated_host

    def random_subnet(self, addr_type=None) -> IPv6SubnetConfig:
        """
        Generates a random IPv6 subnet configuration based on the specified address type.

        Parameters:
        addr_type (Any): The address block type to use for generating the random IPv6 subnet.

        Returns:
        IPv6SubnetConfig: A randomly generated IPv6 subnet configuration.
        """
        subnet = self._random_subnet_for_type(addr_type)
        generated_subnet = IPv6SubnetConfigRandomizer(subnet).randomize()
        return generated_subnet

    def random_hosts_batch(self, addr_type=None, num_ips = 10) -> List[IPv6HostConfig]:
        """
        Generates a batch of random IPv6 host configurations using multiprocessing.

        Parameters:
        addr_type (Any): The address block type to use for generating the random IPv6 hosts.
        num_ips (int): The number of random IPv6 hosts to generate in the batch.

        Returns:
        List[IPv6HostConfig]: A list of randomly generated IPv6 host configurations.
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu_count()) as executor:
            futures = [executor.submit(self.random_host, addr_type) for _ in range(num_ips)]
            random_ips = [future.result() for future in concurrent.futures.as_completed(futures)]
        return random_ips

    def random_subnets_batch(self, addr_type=None, num_ips = 10) -> List[IPv6SubnetConfig]:
        """
        Generates a batch of random IPv6 subnet configurations using multiprocessing.

        Parameters:
        addr_type (Any): The address block type to use for generating the random IPv6 subnets.
        num_ips (int): The number of random IPv6 subnets to generate in the batch.

        Returns:
        List[IPv6SubnetConfig]: A list of randomly generated IPv6 subnet configurations.
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu_count()) as executor:
            futures = [executor.submit(self.random_subnet, addr_type) for _ in range(num_ips)]
            random_subnets = [future.result() for future in concurrent.futures.as_completed(futures)]
        return random_subnets

    def _random_subnet_for_type(self, addr_type) -> IPv6SubnetConfig:
        """
        Retrieves or generates a random IPv6 subnet configuration for the given address type, utilizing caching for performance.

        Parameters:
        addr_type (Any): The address block type to use for generating the IPv6 subnet.

        Returns:
        IPv6SubnetConfig: A random IPv6 subnet configuration.
        """
        if addr_type in self._subnet_cache:
            subnets = self._subnet_cache[addr_type]
        elif addr_type is None:
            subnets = self.batch_subnets(['::/0'])
        else:
            subnets = self.batch_subnets(addr_type.value)
            self._subnet_cache[addr_type] = subnets
        return random.choice(subnets)

class IPConfigRandomizer(ABC):
    """
    Abstract base class for randomizing IP configurations.
    This class provides the structure for generating random IP addresses and netmask configurations based on a given subnet.

    Attributes:
    subnet (IPSubnetConfig): The subnet configuration used as the basis for randomizing IP configurations.
    """
    def __init__(self, subnet):
        """
        Initializes the IPConfigRandomizer with the given subnet.

        Parameters:
        subnet (IPSubnetConfig): The subnet configuration that provides the IP address and netmask for randomization.
        """
        self.subnet = subnet

    @abstractmethod
    def randomize(self) -> InterfaceIPConfig:
        """
        Abstract method to generate a random IP configuration.

        This method must be implemented by subclasses to produce a specific type of random IP configuration (e.g., host, subnet).

        Returns:
        InterfaceIPConfig: A randomly generated IP configuration based on the implementation.
        """
        pass

    def _prepare(self,
                 longest_mask_size: int,
                 decimal_convert_handler: Union[DecimalIPv4ConverterHandler, DecimalIPv6ConverterHandler]
                 ) -> Tuple[IPAddr, IPAddr]:
        """
        Prepares and generates a random IP address and netmask based on the given subnet configuration.

        This helper method handles the bitwise operations needed to create a randomized IP address within the given subnet
        and a random netmask, and it returns both the IP address and the mask in the correct format.

        Parameters:
        longest_mask_size (int): The maximum possible mask size (e.g., 32 for IPv4 or 128 for IPv6).
        decimal_convert_handler (Union[DecimalIPv4ConverterHandler, DecimalIPv6ConverterHandler]):
            The handler used to convert the decimal IP and mask values into their respective binary representations.

        Returns:
        Tuple[IPAddr, IPAddr]: A tuple containing the randomly generated IP address and the corresponding netmask.
        """
        # Get original mask size from the subnet configuration
        original_mask_size = self.subnet.mask.get_mask_size()

        # Choose a random mask size between the original and the longest mask size (e.g., /32 for IPv4)
        random_mask_size = random.randint(original_mask_size, longest_mask_size)

        # Create a mask with all 1s for the chosen random mask size
        ones_for_mask = (1 << random_mask_size) - 1

        # Shift the mask to align it correctly with the number of bits (longest_mask_size - random_mask_size)
        bitwise_random_mask = ones_for_mask << (longest_mask_size - random_mask_size)

        # Calculate the number of host bits by subtracting the original mask size from the total mask size
        host_bits = longest_mask_size - original_mask_size

        # Generate random host bits within the given subnet
        random_host = random.getrandbits(host_bits)

        # Combine the original subnet IP with the randomly generated host bits using a bitwise OR operation
        bitwise_random_host = self.subnet.ip_addr.decimal | random_host

        # Convert the decimal IP and mask to binary using the given handler
        random_host = decimal_convert_handler.handle(bitwise_random_host)
        random_mask = decimal_convert_handler.handle(bitwise_random_mask)

        return random_host, random_mask

class IPv4ConfigRandomizer(IPConfigRandomizer):
    """
    Concrete implementation of IPConfigRandomizer for generating random IPv4 configurations.

    This class provides the implementation for the randomize method to generate random IPv4 addresses and netmasks.
    """
    @abstractmethod
    def randomize(self) -> InterfaceIPConfig:
        """
        Generates a random IPv4 configuration based on the given subnet.

        This method must be implemented by subclasses to produce a specific random IPv4 configuration, such as a host or subnet.

        Returns:
        InterfaceIPConfig: A randomly generated IPv4 configuration.
        """
        pass

class IPv4HostConfigRandomizer(IPv4ConfigRandomizer):
    """
    A concrete implementation of IPv4ConfigRandomizer that generates random IPv4 host configurations.

    This class randomizes an IPv4 host address and its corresponding netmask based on a given subnet configuration.
    It utilizes the `DecimalIPv4ConverterHandler` to convert the randomized IP and netmask into the proper binary format.
    """
    def randomize(self) -> IPv4HostConfig:
        """
        Randomizes and generates an IPv4 host configuration.

        This method uses the `_prepare` method to generate a random host IP address and netmask within a given subnet.
        The randomization is based on the longest mask size (32 for IPv4) and converts the IP and netmask using `DecimalIPv4ConverterHandler`.

        Returns:
        IPv4HostConfig: A randomly generated host configuration containing the IP address and the netmask.
        """
        random_host, random_mask = self._prepare(32, DecimalIPv4ConverterHandler())
        ipv4_host = IPv4HostConfig(IPv4Addr(random_host), IPv4NetMask(random_mask))
        return ipv4_host

class IPv4SubnetConfigRandomizer(IPv4ConfigRandomizer):
    """
    A concrete implementation of IPv4ConfigRandomizer that generates random IPv4 subnet configurations.

    This class randomizes an IPv4 subnet address and its corresponding netmask based on a given subnet configuration.
    It utilizes the `DecimalIPv4ConverterHandler` to convert the randomized IP and netmask into the proper binary format.
    """
    def randomize(self) -> IPv4SubnetConfig:
        """
        Randomizes and generates an IPv4 subnet configuration.

        This method uses the `_prepare` method to generate a random subnet IP address and netmask within a given subnet.
        The randomization is based on the longest mask size (32 for IPv4) and converts the IP and netmask using `DecimalIPv4ConverterHandler`.

        Returns:
        IPv4SubnetConfig: A randomly generated subnet configuration containing the IP address and the netmask.
        """
        random_host, random_mask = self._prepare(32, DecimalIPv4ConverterHandler())
        ipv4_subnet = IPv4SubnetConfig(IPv4Addr(random_host), IPv4NetMask(random_mask))
        return ipv4_subnet

class IPv6ConfigRandomizer(IPConfigRandomizer):
    """
    Abstract class for generating random IPv6 configurations.

    This class provides the structure for generating random IPv6 addresses and netmask configurations based on a given subnet.
    Subclasses must implement the `randomize` method to generate specific IPv6 configurations (e.g., host, subnet).
    """
    @abstractmethod
    def randomize(self) -> InterfaceIPConfig:
        """
        Abstract method to generate a random IPv6 configuration.

        This method must be implemented by subclasses to produce a specific random IPv6 configuration (such as a host or subnet).

        Returns:
        InterfaceIPConfig: A randomly generated IPv6 configuration.
        """
        pass

class IPv6HostConfigRandomizer(IPv6ConfigRandomizer):
    """
    A concrete implementation of IPv6ConfigRandomizer that generates random IPv6 host configurations.

    This class randomizes an IPv6 host address and its corresponding netmask based on a given subnet configuration.
    It uses the `DecimalIPv6ConverterHandler` to convert the randomized IP and netmask into their binary format.
    """
    def randomize(self) -> IPv6HostConfig:
        """
        Randomizes and generates an IPv6 host configuration.

        This method uses the `_prepare` method to generate a random host IP address and netmask within a given IPv6 subnet.
        The randomization is based on the longest mask size (128 for IPv6) and converts the IP and netmask using `DecimalIPv6ConverterHandler`.

        Returns:
        IPv6HostConfig: A randomly generated IPv6 host configuration containing the IP address and the netmask.
        """
        random_host, random_mask = self._prepare(128, DecimalIPv6ConverterHandler())
        ipv6_host = IPv6HostConfig(IPv6Addr(random_host), IPv6NetMask(random_mask))
        return ipv6_host

class IPv6SubnetConfigRandomizer(IPv6ConfigRandomizer):
    """
    A concrete implementation of IPv6ConfigRandomizer that generates random IPv6 subnet configurations.

    This class randomizes an IPv6 subnet address and its corresponding netmask based on a given subnet configuration.
    It uses the `DecimalIPv6ConverterHandler` to convert the randomized IP and netmask into their binary format.
    """
    def randomize(self) -> IPv6SubnetConfig:
        """
        Randomizes and generates an IPv6 subnet configuration.

        This method uses the `_prepare` method to generate a random subnet IP address and netmask within a given IPv6 subnet.
        The randomization is based on the longest mask size (128 for IPv6) and converts the IP and netmask using `DecimalIPv6ConverterHandler`.

        Returns:
        IPv6SubnetConfig: A randomly generated IPv6 subnet configuration containing the IP address and the netmask.
        """
        random_host, random_mask = self._prepare(128, DecimalIPv6ConverterHandler())
        ipv6_subnet = IPv6SubnetConfig(IPv6Addr(random_host), IPv6NetMask(random_mask))
        return ipv6_subnet



from enum import Enum
from typing import Dict, List, Union

from ttlinks.common.binary_utils.binary import Octet
from ttlinks.common.binary_utils.binary_factory import OctetFlyWeightFactory
from ttlinks.common.design_template.factory import Factory
from ttlinks.common.tools.converters import NumeralConverter
from ttlinks.macservice.mac_converters import MACConverter


class OUIType(Enum):
    """
    An enumeration to categorize different types of Organizationally Unique Identifiers (OUIs).
    OUI types include:
    UNKNOWN: Unidentified or unspecified OUI type.
    IAB: Individual Address Block for smaller requirements.
    MA_S: Standard MAC Address block size typically used by most organizations.
    MA_M: Medium-sized MAC Address block for larger organizational needs.
    MA_L: Large MAC Address block allocated for extensive hardware requirements.
    CID: Company ID, often used to identify specific entities or manufacturers beyond the usual MAC address allocations.
    """
    UNKNOWN = 0
    IAB = 1
    MA_S = 2
    MA_M = 3
    MA_L = 4
    CID = 5


class OUIUnit:
    """
    Represents an Organizational Unique Identifier (OUI) unit that is used for identifying the manufacturer
    or organization by a unique MAC prefix in network technologies. This class utilizes the flyweight design pattern
    to manage instances efficiently by ensuring that identical objects are shared rather than duplicated.

    Attributes:
        _oui_units (dict): A class-level dictionary that stores unique OUIUnit instances indexed by a tuple key.
    """
    _oui_units = {}

    def __new__(
            cls,
            oui_id: List[Octet],
            oui_mask: List[Octet],
            oui_type: OUIType,
            organization: Union[str, None],
            mac_range: Union[str, None],
            oui_hex: Union[str, None],
            address: Union[str, None]
    ):
        """
        Ensures that only one instance of OUIUnit is created for each unique combination of start binaries,
        MAC mask, and OUI type. If an instance with the same key already exists, it returns the existing instance
        instead of creating a new one.

        Parameters:
            oui_id (List[BinaryClass]): Binary representations of the start of the OUI.
            oui_mask (List[BinaryClass]): Binary mask that applies to the OUI.
            oui_type (OUIType): The type of the OUI, indicating its use.
            organization (Union[str, None]): Name of the organization associated with the OUI.
            mac_range (Union[str, None]): The MAC range associated with the OUI.
            oui_hex (Union[str, None]): The hexadecimal representation of the OUI.
            address (Union[str, None]): The address of the organization owning the OUI.

        Returns:
            OUIUnit: A new or existing instance of the OUIUnit.
        """
        key = (tuple(oui_id), tuple(oui_mask), oui_type)
        if key not in cls._oui_units.keys():
            instance = super().__new__(cls)
            cls._oui_units[key] = instance
            instance.__oui_id = oui_id
            instance.__oui_mask = oui_mask
            instance.__oui_type = oui_type
            instance.__organization = organization
            instance.__mac_range = mac_range
            instance.__oui_hex = oui_hex
            instance.__address = address
            return instance
        return cls._oui_units[key]

    def __init__(
            self,
            oui_id: List[Octet],
            oui_mask: List[Octet],
            oui_type: OUIType,
            organization: Union[str, None],
            mac_range: Union[str, None],
            oui_hex: Union[str, None],
            address: Union[str, None],
    ):
        """
        Initializes the OUIUnit instance with provided parameters. This method is typically called only
        once when the instance is first created.

        Parameters:
            oui_id, oui_mask, oui_type, organization, mac_range, oui_hex, address: See __new__ for details.
        """
        self.__oui_id = oui_id
        self.__oui_mask = oui_mask
        self.__oui_type = oui_type
        self.__organization = organization
        self.__mac_range = mac_range
        self.__oui_hex = oui_hex
        self.__address = address

    @property
    def oui_id_binary_digits(self) -> List[int]:
        """
        Returns the binary digits for the OUI's identifier as a list of integers.

        Returns:
        - List[int]: Binary digits of the OUI ID.
        """
        result = []
        for octet in self.__oui_id:
            result.extend(octet.binary_digits)
        return result

    @property
    def oui_mask_binary_digits(self) -> List[int]:
        """
        Returns the binary digits for the OUI's mask as a list of integers.

        Returns:
        - List[int]: Binary digits of the OUI mask.
        """
        result = []
        for octet in self.__oui_mask:
            result.extend(octet.binary_digits)
        return result

    @property
    def record(self) -> Dict:
        """
        Returns a dictionary containing the hexadecimal representation of the OUI details.

        Returns:
        - Dict: A dictionary with keys 'oui_id', 'oui_mask', 'oui_type', 'organization', 'mac_range', 'oui_hex', 'address'.
        """
        return {
            'oui_id': ':'.join([
                NumeralConverter.binary_to_hexadecimal(str(start_bin)).rjust(2, '0')
                for start_bin in self.__oui_id
            ]),
            'oui_mask': ':'.join([
                NumeralConverter.binary_to_hexadecimal(str(mask_bin)).rjust(2, '0')
                for mask_bin in self.__oui_mask
            ]),
            'oui_type': self.__oui_type.name,
            'organization': self.__organization,
            'mac_range': self.__mac_range,
            'oui_hex': self.__oui_hex,
            'address': self.__address
        }


class OUIUnitCreator(Factory):
    """
    A factory class for creating instances of OUIUnit. It processes the raw input and creates the appropriate OUIUnit.

    Method:
    - create_product: Creates a new OUIUnit based on the provided arguments (oui_id, oui_mask, oui_type, etc.).
    """

    def create_product(self, **kwargs):
        """
        Converts the provided OUI data to octets and creates an OUIUnit instance.

        Parameters:
        - kwargs (dict): Contains raw input like 'oui_id', 'oui_mask', 'oui_type', etc., which is processed before creating the OUIUnit.

        Returns:
        - OUIUnit: A new instance of the OUIUnit class.
        """
        data = {
            'oui_id': MACConverter.convert_oui(kwargs['oui_id']),
            'oui_mask': MACConverter.convert_oui(kwargs['oui_mask']),
            'oui_type': OUIType[kwargs['oui_type']]
        }
        kwargs.update(data)
        return OUIUnit(**kwargs)


class OUIDBStrategy(Enum):
    """
    An enumeration to categorize different strategies for loading or searching OUI data.
    OUI loader strategies include:
    - SIMPLE_ITERATION: A simple iteration strategy for loading or searching OUI data.
    - TRIE: A trie-based strategy for loading or searching OUI data.
    """
    SIMPLE_ITERATION = 0
    TRIE = 1

class OUIMask(Enum):
    """
    An enumeration to categorize different types of masks used for OUI search.
    OUI masks include:
    """
    IAB = (
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('FF'))] * 4 +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('F0'))] +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('00'))]
    )
    MA_S = (
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('FF'))] * 4 +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('F0'))] +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('00'))]
    )
    MA_M = (
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('FF'))] * 3 +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('F0'))] +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('00'))] * 2
    )
    MA_L = (
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('FF'))] * 3 +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('00'))] * 3
    )
    CID = (
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('FF'))] * 3 +
            [OctetFlyWeightFactory.get_octet(NumeralConverter.hexadecimal_to_binary('00'))] * 3
    )



import socket
import time

# 提取并转换 IP 地址
source_ip_hex = 'c0a80146'  # 192.168.1.70
destination_ip_hex = '12dcb641'  # 18.220.182.65

source_ip = socket.inet_ntoa(bytes.fromhex(source_ip_hex))
destination_ip = socket.inet_ntoa(bytes.fromhex(destination_ip_hex))

print("源 IP 地址：", source_ip)
print("目的 IP 地址：", destination_ip)

import struct

# 构建伪首部
source_ip_bytes = socket.inet_aton(source_ip)  # '192.168.1.70'
destination_ip_bytes = socket.inet_aton(destination_ip)  # '18.220.182.65'
reserved = b'\x00'
protocol = b'\x06'  # TCP 协议号为 6

# 计算 TCP 长度（TCP 头 + 数据）
tcp_segment_hex = 'd38c01bb51aed728000000008002faf000000000020405b40103030801010402'
tcp_segment_bytes = bytes.fromhex(tcp_segment_hex)
tcp_length = len(tcp_segment_bytes)
tcp_length_bytes = struct.pack('!H', tcp_length)  # 以网络字节序打包为 2 字节

# 组合伪首部
pseudo_header = source_ip_bytes + destination_ip_bytes + reserved + protocol + tcp_length_bytes

print("伪首部（hex）：", pseudo_header.hex())

# 校验和计算的数据
checksum_data = pseudo_header + tcp_segment_bytes

print("用于校验和计算的数据（hex）：", checksum_data.hex())


def calculate_checksum(data):
    # 如果数据长度为奇数，填充一个字节
    if len(data) % 2 == 1:
        data += b'\x00'

    checksum = 0
    # 以每 2 个字节为一组进行求和
    for i in range(0, len(data), 2):
        word = (data[i] << 8) + data[i + 1]
        checksum += word
        # 处理可能的进位
        checksum = (checksum & 0xFFFF) + (checksum >> 16)

    # 对最终的和取反
    checksum = ~checksum & 0xFFFF
    return checksum


calculated_checksum = calculate_checksum(checksum_data)
print("计算得到的校验和：0x{:04x}".format(calculated_checksum))

time.sleep(5)


import pytest
from ttlinks.common.tools.converters import NumeralConverter

def test_binary_to_decimal():
    # Test converting binary to decimal
    assert NumeralConverter.binary_to_decimal("101010") == 42
    assert NumeralConverter.binary_to_decimal("1111") == 15

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.binary_to_decimal(101010)  # Not a string

def test_decimal_to_binary():
    # Test converting decimal to binary
    assert NumeralConverter.decimal_to_binary(42) == "00101010"
    assert NumeralConverter.decimal_to_binary(15, r_just=4) == "1111"

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.decimal_to_binary("42")  # Not an integer

def test_binary_to_hexadecimal():
    # Test converting binary to hexadecimal
    assert NumeralConverter.binary_to_hexadecimal("101010") == "2A"
    assert NumeralConverter.binary_to_hexadecimal("1111") == "F"

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.binary_to_hexadecimal(101010)  # Not a string

def test_hexadecimal_to_binary():
    # Test converting hexadecimal to binary
    assert NumeralConverter.hexadecimal_to_binary("2A") == "00101010"
    assert NumeralConverter.hexadecimal_to_binary("F", r_just=4) == "1111"

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.hexadecimal_to_binary(42)  # Not a string
    with pytest.raises(TypeError):
        NumeralConverter.hexadecimal_to_binary("G2")  # Invalid hexadecimal format

def test_hexadecimal_to_decimal():
    # Test converting hexadecimal to decimal
    assert NumeralConverter.hexadecimal_to_decimal("2A") == 42
    assert NumeralConverter.hexadecimal_to_decimal("F") == 15

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.hexadecimal_to_decimal(42)  # Not a string
    with pytest.raises(TypeError):
        NumeralConverter.hexadecimal_to_decimal("G2")  # Invalid hexadecimal format

def test_decimal_to_hexadecimal():
    # Test converting decimal to hexadecimal
    assert NumeralConverter.decimal_to_hexadecimal(42) == "2A"
    assert NumeralConverter.decimal_to_hexadecimal(15, r_just=2) == "0F"

    # Test invalid input
    with pytest.raises(TypeError):
        NumeralConverter.decimal_to_hexadecimal("42")  # Not an integer



import pytest
from ttlinks.common.binary_utils.binary_factory import OctetFlyWeightFactory
from ttlinks.ipservice.ip_converters import (
    OctetIPv4ConverterHandler, BinaryDigitsIPv4ConverterHandler, CIDRIPv4ConverterHandler,
    DotIPv4ConverterHandler, OctetIPv6ConverterHandler, BinaryDigitsIPv6ConverterHandler,
    CIDRIPv6ConverterHandler, ColonIPv6ConverterHandler
)


# IPv4 Octet Handler Tests
def test_octet_ipv4_handler_valid_address():
    request = [
        OctetFlyWeightFactory.get_octet('11000000'),  # 192
        OctetFlyWeightFactory.get_octet('10101000'),  # 168
        OctetFlyWeightFactory.get_octet('00000001'),  # 1
        OctetFlyWeightFactory.get_octet('00000001')  # 1
    ]
    result = OctetIPv4ConverterHandler().handle(request)
    assert result == request, "Should return the input octets for a valid IPv4 address."


def test_octet_ipv4_handler_less_octets():
    request = [
        OctetFlyWeightFactory.get_octet('11000000'),  # 192
        OctetFlyWeightFactory.get_octet('10101000')  # 168
    ]
    result = OctetIPv4ConverterHandler().handle(request)
    assert result is None, "Should return None for invalid IPv4 address with less than 4 octets."


def test_octet_ipv4_handler_more_octets():
    request = [
        OctetFlyWeightFactory.get_octet('11000000'),  # 192
        OctetFlyWeightFactory.get_octet('10101000'),  # 168
        OctetFlyWeightFactory.get_octet('00000001'),  # 1
        OctetFlyWeightFactory.get_octet('00000001'),  # 1
        OctetFlyWeightFactory.get_octet('00000010')  # 2
    ]
    result = OctetIPv4ConverterHandler().handle(request)
    assert result is None, "Should return None for invalid IPv4 address with more than 4 octets."


def test_octet_ipv4_handler_empty_list():
    request = []
    result = OctetIPv4ConverterHandler().handle(request)
    assert result is None, "Should return None for an empty list of octets."


def test_octet_ipv4_handler_none_input():
    result = OctetIPv4ConverterHandler().handle(None)
    assert result is None, "Should return None when None is passed as input."


def test_octet_ipv4_handler_non_octet_list():
    request = [192, 168, 1, 1]  # Incorrect types, should be instances of Octet
    result = OctetIPv4ConverterHandler().handle(request)
    assert result is None, "Should return None for non-octet type inputs."


# IPv4 Binary Digits Handler Tests
def test_binary_digits_ipv4_handler_valid_address():
    request = [
        1, 1, 0, 0, 0, 0, 0, 0,  # 192
        1, 0, 1, 0, 1, 0, 0, 0,  # 168
        0, 0, 0, 0, 0, 0, 0, 1,  # 1
        0, 0, 0, 0, 0, 0, 0, 1  # 1
    ]
    expected_octets = [
        OctetFlyWeightFactory.get_octet('11000000'),
        OctetFlyWeightFactory.get_octet('10101000'),
        OctetFlyWeightFactory.get_octet('00000001'),
        OctetFlyWeightFactory.get_octet('00000001')
    ]
    converter = BinaryDigitsIPv4ConverterHandler()
    result = converter.handle(request)
    assert result == expected_octets, "Should correctly convert binary digits to octets for a valid IPv4 address."


def test_binary_digits_ipv4_handler_incorrect_number_of_digits():
    request_short = [1, 1, 0, 0]  # Too few digits
    request_long = [1] * 33  # Too many digits
    converter = BinaryDigitsIPv4ConverterHandler()
    assert converter.handle(request_short) is None, "Should return None for too few binary digits."
    assert converter.handle(request_long) is None, "Should return None for too many binary digits."


def test_binary_digits_ipv4_handler_invalid_digit_values():
    request = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 1, 0, 1, 0, 1, 1, 0, 2, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0]
    converter = BinaryDigitsIPv4ConverterHandler()
    result = converter.handle(request)
    assert result is None, "Should return None for invalid binary values."


def test_binary_digits_ipv4_handler_empty_list():
    request = []
    converter = BinaryDigitsIPv4ConverterHandler()
    result = converter.handle(request)
    assert result is None, "Should return None for an empty list of binary digits."


def test_binary_digits_ipv4_handler_none_input():
    converter = BinaryDigitsIPv4ConverterHandler()
    result = converter.handle(None)
    assert result is None, "Should return None when None is passed as input."


# IPv4 CIDR Handler Tests
def test_cidr_ipv4_handler_valid_notation():
    converter = CIDRIPv4ConverterHandler()
    test_cases = [
        ("/24", [255, 255, 255, 0]),
        ("/16", [255, 255, 0, 0]),
        ("/8", [255, 0, 0, 0]),
        ("/32", [255, 255, 255, 255]),
        ("/0", [0, 0, 0, 0])
    ]
    for cidr, expected in test_cases:
        expected_octets = [OctetFlyWeightFactory.get_octet(format(num, '08b')) for num in expected]
        result = converter.handle(cidr)
        assert result == expected_octets, f"Failed for CIDR {cidr}. Expected {expected_octets}, got {result}"


def test_cidr_ipv4_handler_invalid_notation():
    invalid_cidrs = ["/33", "/-1", "/abc", "192.168.1.0/24", "/"]
    converter = CIDRIPv4ConverterHandler()
    for cidr in invalid_cidrs:
        result = converter.handle(cidr)
        assert result is None, f"Should return None for invalid CIDR {cidr}"


def test_cidr_ipv4_handler_non_string_input():
    non_strings = [None, 24, [], {}, 192168010]
    converter = CIDRIPv4ConverterHandler()
    for input_val in non_strings:
        result = converter.handle(input_val)
        assert result is None, f"Should return None for non-string input {input_val}"


def test_cidr_ipv4_handler_empty_string():
    converter = CIDRIPv4ConverterHandler()
    result = converter.handle("")
    assert result is None, "Should return None for an empty string input"


# IPv4 Dot Handler Tests
def test_dot_ipv4_handler_valid_address():
    converter = DotIPv4ConverterHandler()
    ipv4_address = "192.168.1.1"
    expected = [
        OctetFlyWeightFactory.get_octet('11000000'),  # 192
        OctetFlyWeightFactory.get_octet('10101000'),  # 168
        OctetFlyWeightFactory.get_octet('00000001'),  # 1
        OctetFlyWeightFactory.get_octet('00000001')  # 1
    ]
    result = converter.handle(ipv4_address)
    assert result == expected, f"Should correctly convert {ipv4_address} to octets."


def test_dot_ipv4_handler_invalid_address():
    converter = DotIPv4ConverterHandler()
    invalid_addresses = ["192.168.256.1", "10.10.10", "10.10", "300.300.300.300", "192.168.1.1.1"]
    for address in invalid_addresses:
        result = converter.handle(address)
        assert result is None, f"Should return None for invalid IPv4 address {address}"


def test_dot_ipv4_handler_non_string_input():
    converter = DotIPv4ConverterHandler()
    non_strings = [None, 12345, 19216801, ["192", "168", "1", "1"], {"ip": "192.168.1.1"}]
    for input_val in non_strings:
        result = converter.handle(input_val)
        assert result is None, f"Should return None for non-string input {input_val}"


def test_dot_ipv4_handler_empty_string():
    converter = DotIPv4ConverterHandler()
    result = converter.handle("")
    assert result is None, "Should return None for an empty string input"


# IPv6 Octet Handler Tests
def test_octet_ipv6_handler_valid_address():
    converter = OctetIPv6ConverterHandler()
    request = [OctetFlyWeightFactory.get_octet('00100000') for _ in range(16)]  # Simplified for brevity
    result = converter.handle(request)
    assert result == request, "Should return the input octets for a valid IPv6 address."


def test_octet_ipv6_handler_less_octets():
    converter = OctetIPv6ConverterHandler()
    request = [OctetFlyWeightFactory.get_octet('00100000') for _ in range(15)]  # 15 octets
    result = converter.handle(request)
    assert result is None, "Should return None for invalid IPv6 address with less than 16 octets."


def test_octet_ipv6_handler_more_octets():
    converter = OctetIPv6ConverterHandler()
    request = [OctetFlyWeightFactory.get_octet('00100000') for _ in range(17)]  # 17 octets
    result = converter.handle(request)
    assert result is None, "Should return None for invalid IPv6 address with more than 16 octets."


def test_octet_ipv6_handler_empty_list():
    converter = OctetIPv6ConverterHandler()
    result = converter.handle([])
    assert result is None, "Should return None for an empty list of octets."


def test_octet_ipv6_handler_none_input():
    converter = OctetIPv6ConverterHandler()
    result = converter.handle(None)
    assert result is None, "Should return None when None is passed as input."


def test_octet_ipv6_handler_non_octet_list():
    converter = OctetIPv6ConverterHandler()
    request = ['00100000'] * 16  # Incorrect input as strings instead of octet objects
    result = converter.handle(request)
    assert result is None, "Should return None for non-octet type inputs."


# IPv6 Binary Digits Handler Tests
def test_binary_digits_ipv6_handler_valid_address():
    converter = BinaryDigitsIPv6ConverterHandler()
    request = [0] * 64 + [1] * 64  # Simplified example of ::FFFF:FFFF:FFFF:FFFF
    expected_octets = [OctetFlyWeightFactory.get_octet('00000000')] * 8 + \
                      [OctetFlyWeightFactory.get_octet('11111111')] * 8
    result = converter.handle(request)
    assert result == expected_octets, "Should correctly convert 128 binary digits to octets for a valid IPv6 address."


def test_binary_digits_ipv6_handler_incorrect_number_of_digits():
    converter = BinaryDigitsIPv6ConverterHandler()
    request_short = [0] * 100  # Too few digits
    request_long = [0] * 140  # Too many digits
    assert converter.handle(request_short) is None, "Should return None for too few binary digits."
    assert converter.handle(request_long) is None, "Should return None for too many binary digits."


def test_binary_digits_ipv6_handler_invalid_digit_values():
    converter = BinaryDigitsIPv6ConverterHandler()
    request_invalid = [2] * 128  # Invalid binary digit values
    result = converter.handle(request_invalid)
    assert result is None, "Should return None for invalid binary values."


def test_binary_digits_ipv6_handler_empty_list():
    converter = BinaryDigitsIPv6ConverterHandler()
    result = converter.handle([])
    assert result is None, "Should return None for an empty list of binary digits."


def test_binary_digits_ipv6_handler_none_input():
    converter = BinaryDigitsIPv6ConverterHandler()
    result = converter.handle(None)
    assert result is None, "Should return None when None is passed as input."


def test_binary_digits_ipv6_handler_non_integer_list():
    converter = BinaryDigitsIPv6ConverterHandler()
    request = ['0', '1', '0', '1'] * 32  # Non-integer inputs
    result = converter.handle(request)
    assert result is None, "Should return None for non-integer input items."


# IPv6 CIDR Handler Tests
def test_cidr_ipv6_handler_valid_notation():
    converter = CIDRIPv6ConverterHandler()
    test_cases = [
        ("/64", ['11111111'] * 8 + ['00000000'] * 8),
        ("/128", ['11111111'] * 16),
        ("/0", ['00000000'] * 16),
        ("/32", ['11111111'] * 4 + ['00000000'] * 12),
        ("/114", ['11111111'] * 14 + ['11000000', '00000000'])
    ]
    for cidr, expected in test_cases:
        expected_octets = [OctetFlyWeightFactory.get_octet(binary_string) for binary_string in expected]
        result = converter.handle(cidr)
        assert result == expected_octets, f"Failed for CIDR {cidr}. Expected {expected_octets}, got {result}"


def test_cidr_ipv6_handler_invalid_notation():
    converter = CIDRIPv6ConverterHandler()
    invalid_cidrs = ["/129", "/-1", "192.168.1.0/24", "/abc", "", "/ "]
    for cidr in invalid_cidrs:
        result = converter.handle(cidr)
        assert result is None, f"Should return None for invalid CIDR {cidr}"


def test_cidr_ipv6_handler_non_string_input():
    converter = CIDRIPv6ConverterHandler()
    non_strings = [None, 24, [], {}, 19216801]
    for input_val in non_strings:
        result = converter.handle(input_val)
        assert result is None, f"Should return None for non-string input {input_val}"


def test_cidr_ipv6_handler_empty_string():
    converter = CIDRIPv6ConverterHandler()
    result = converter.handle("")
    assert result is None, "Should return None for an empty string input"


# IPv6 Colon Handler Tests
def test_colon_ipv6_handler_valid_address():
    converter = ColonIPv6ConverterHandler()
    test_cases = [
        ("2001:0db8:85a3:0000:0000:8a2e:0370:7334",
         ["00100000", "00000001", "00001101", "10111000", "10000101", "10100011", "00000000", "00000000",
          "00000000", "00000000", "10001010", "00101110", "00000011", "01110000", "01110011", "00110100"]),
        ("2001:db8::1",
         ["00100000", "00000001", "00001101", "10111000", "00000000", "00000000", "00000000", "00000000",
          "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000000", "00000001"]),
        ("::1",
         ["00000000"] * 15 + ["00000001"]),
        ("::",
         ["00000000"] * 16)
    ]
    for ipv6, expected_binaries in test_cases:
        expected_octets = [OctetFlyWeightFactory.get_octet(binary) for binary in expected_binaries]
        result = converter.handle(ipv6)
        assert result == expected_octets, f"Failed for IPv6 {ipv6}. Expected {expected_octets}, got {result}"


def test_colon_ipv6_handler_invalid_address():
    converter = ColonIPv6ConverterHandler()
    invalid_addresses = ["2001:0db8:85a3:0000:0000:8a2e:0370:7334:8765",  # Too many segments
                         "2001:db8::85a3::7334",  # Double "::"
                         "2001:0dbg:85a3::7334",  # Invalid hex character
                         "2001:db8::85a3:"]  # Incomplete segment
    for address in invalid_addresses:
        result = converter.handle(address)
        assert result is None, f"Should return None for invalid IPv6 address {address}"


def test_colon_ipv6_handler_non_string_input():
    converter = ColonIPv6ConverterHandler()
    non_strings = [None, 12345, 19216801, ["2001", "0db8", "85a3"], {"ip": "2001:db8::1"}]
    for input_val in non_strings:
        result = converter.handle(input_val)
        assert result is None, f"Should return None for non-string input {input_val}"


def test_colon_ipv6_handler_empty_string():
    converter = ColonIPv6ConverterHandler()
    result = converter.handle("")
    assert result is None, "Should return None for an empty string input"


# IPv4 Binary Digits Handler Tests
def test_binary_digits_ipv4_handler_boundary_values():
    converter = BinaryDigitsIPv4ConverterHandler()
    # Test with exactly 32 binary digits
    request = [1] * 32
    result = converter.handle(request)
    assert result is not None, "Should handle exactly 32 binary digits."


# IPv4 CIDR Handler Tests
def test_cidr_ipv4_handler_boundary_values():
    converter = CIDRIPv4ConverterHandler()
    # Test with boundary CIDR values
    assert converter.handle("/0") is not None, "Should handle /0 CIDR."
    assert converter.handle("/32") is not None, "Should handle /32 CIDR."


# IPv6 Binary Digits Handler Tests
def test_binary_digits_ipv6_handler_boundary_values():
    converter = BinaryDigitsIPv6ConverterHandler()
    # Test with exactly 128 binary digits
    request = [1] * 128
    result = converter.handle(request)
    assert result is not None, "Should handle exactly 128 binary digits."


# IPv6 CIDR Handler Tests
def test_cidr_ipv6_handler_boundary_values():
    converter = CIDRIPv6ConverterHandler()
    # Test with boundary CIDR values
    assert converter.handle("/0") is not None, "Should handle /0 CIDR."
    assert converter.handle("/128") is not None, "Should handle /128 CIDR."


# IPv4 Dot Handler Tests
def test_dot_ipv4_handler_leading_zeros():
    converter = DotIPv4ConverterHandler()
    ipv4_address = "192.168.001.001"
    expected = [
        OctetFlyWeightFactory.get_octet('11000000'),  # 192
        OctetFlyWeightFactory.get_octet('10101000'),  # 168
        OctetFlyWeightFactory.get_octet('00000001'),  # 1
        OctetFlyWeightFactory.get_octet('00000001')  # 1
    ]
    result = converter.handle(ipv4_address)
    assert result == expected, f"Should correctly convert {ipv4_address} to octets."


# IPv6 Colon Handler Tests
def test_colon_ipv6_handler_leading_zeros():
    converter = ColonIPv6ConverterHandler()
    ipv6_address = "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    expected_binaries = [
        "00100000", "00000001", "00001101", "10111000", "10000101", "10100011", "00000000", "00000000",
        "00000000", "00000000", "10001010", "00101110", "00000011", "01110000", "01110011", "00110100"
    ]
    expected_octets = [OctetFlyWeightFactory.get_octet(binary) for binary in expected_binaries]
    result = converter.handle(ipv6_address)
    assert result == expected_octets, f"Failed for IPv6 {ipv6_address}. Expected {expected_octets}, got {result}"



import struct
from abc import abstractmethod, ABC
from typing import List, Dict, Any

from ttlinks.common.design_template.cor import SimpleCoRHandler
from ttlinks.protocol_stack.base_classes.header_builder import Header, HeaderBuilder
from ttlinks.protocol_stack.base_classes.stack_utils import ProtocolUnit

# -------------------TCP Option Value Object-------------------
class TCPOptionSACKValue:
    """
    Represents a TCP Selective Acknowledgment (SACK) option value, which specifies a range of bytes
    acknowledged by the sender as received.

    Attributes:
    - left_edge (int): The left (starting) edge of the acknowledged range.
    - right_edge (int): The right (ending) edge of the acknowledged range.

    Methods:
    - _validate: Ensures that left and right edges are valid, non-negative integers within the 32-bit range.
    - __repr__: Provides a string representation of the SACK range.

    Exceptions:
    - Raises ValueError if the edges do not meet the specified constraints.
    """
    def __init__(self, left_edge: int, right_edge: int):
        """
        Initializes the TCPOptionSACKValue with left and right edges, validating them.

        Parameters:
        - left_edge (int): The starting edge of the SACK range.
        - right_edge (int): The ending edge of the SACK range.
        """
        self._validate(left_edge, right_edge)
        self._left_edge = left_edge
        self._right_edge = right_edge

    @staticmethod
    def _validate(left_edge: int, right_edge: int):
        """
        Validates that left and right edges are within acceptable ranges.

        Parameters:
        - left_edge (int): The starting edge to validate.
        - right_edge (int): The ending edge to validate.

        Raises:
        - ValueError: If edges are invalid.
        """
        if left_edge > right_edge:
            raise ValueError('Left edge must be less than or equal to right edge')
        if left_edge < 0 or right_edge < 0:
            raise ValueError('Left and right edges must be greater than or equal to 0')
        if left_edge > 0xFFFFFFFF or right_edge > 0xFFFFFFFF:
            raise ValueError('Left and right edges must be less than or equal to 4294967295')

    @property
    def left_edge(self) -> int:
        """
        Returns the left edge of the SACK range.

        Returns:
        - int: Left edge of the range.
        """
        return self._left_edge

    @property
    def right_edge(self) -> int:
        """
        Returns the right edge of the SACK range.

        Returns:
        - int: Right edge of the range.
        """
        return self._right_edge

    def __repr__(self):
        """
        Provides a string representation of the SACK range.

        Returns:
        - str: Formatted string of the SACK range.
        """
        return f'SACKRange(left_edge={self._left_edge}, right_edge={self._right_edge})'

class TCPOptionTimestampValue:
    """
    Represents a TCP Timestamp option value, including a timestamp and an echo reply for
    round-trip time measurements.

    Attributes:
    - timestamp (int): The current timestamp value.
    - echo_reply (int): Echoed timestamp from the receiving host.

    Methods:
    - __repr__: Provides a string representation of the timestamp and echo reply values.
    """
    def __init__(self, timestamp: int, echo_reply: int):
        """
        Initializes the TCPOptionTimestampValue with a timestamp and echo reply.

        Parameters:
        - timestamp (int): The timestamp value.
        - echo_reply (int): The echoed timestamp value.
        """
        self._timestamp = timestamp
        self._echo_reply = echo_reply

    @property
    def timestamp(self) -> int:
        """
        Returns the timestamp value.

        Returns:
        - int: The timestamp value.
        """
        return self._timestamp

    @property
    def echo_reply(self) -> int:
        """
        Returns the echoed timestamp value.

        Returns:
        - int: The echoed timestamp value.
        """
        return self._echo_reply

    def __repr__(self):
        """
        Provides a string representation of the timestamp and echo reply values.

        Returns:
        - str: Formatted string with timestamp and echo reply.
        """
        return f'Timestamp(timestamp={self._timestamp}, echo_reply={self._echo_reply})'

# -------------------TCP Option Unit-------------------
class TCPOptionUnit(ProtocolUnit, ABC):
    """
    Abstract base class for TCP options, supporting a variety of TCP options with unique identifiers.

    Supported TCP Options:
    - End of Option List
    - No-Operation
    - Maximum Segment Size
    - Window Scale
    - SACK Permitted
    - SACK
    - Timestamp
    - Unknown Option

    Attributes:
    - kind (bytes): Option type identifier.
    - length (bytes): Length of the option in bytes.
    - value (bytes): The option's data payload.

    Properties:
    - as_bytes: Returns the complete option as a byte sequence.
    - attributes: Provides a dictionary of the option’s raw kind, length, and value.
    - summary: Provides a human-readable dictionary of option attributes.

    Methods:
    - kind: Abstract property for the option type.
    - length: Abstract property for the option length.
    - value: Abstract property for the option value.
    """
    def __init__(self, kind: bytes, length: bytes=b'', value: bytes=b''):
        """
        Initializes the TCP option with its kind, length, and value.

        Parameters:
        - kind (bytes): The kind of the TCP option.
        - length (bytes): The length of the option, defaults to an empty byte string.
        - value (bytes): The option's value or data, defaults to an empty byte string.
        """
        self._kind = kind
        self._length = length
        self._value = value

    @property
    def as_bytes(self) -> bytes:
        """
        Returns the TCP option as a concatenated byte sequence.

        Returns:
        - bytes: The TCP option as bytes.
        """
        return self._kind + self._length + self._value
    @property
    def attributes(self)-> dict:
        """
        Returns a dictionary of the TCP option's raw attributes.

        Returns:
        - dict: Dictionary containing the 'kind', 'length', and 'value' attributes.
        """
        return {
            "kind": self._kind,
            "length": self._length,
            "value": self._value,
        }
    @property
    def summary(self) -> dict:
        """
        Returns a human-readable dictionary of the TCP option's attributes.

        Returns:
        - dict: Dictionary with descriptive keys for 'kind', 'length', and 'value'.
        """
        return {
            "kind": self.kind,
            "length": self.length,
            "value": self.value,
        }
    @property
    @abstractmethod
    def kind(self):
        pass
    @property
    @abstractmethod
    def length(self):
        pass
    @property
    @abstractmethod
    def value(self):
        pass
    def __repr__(self):
        """
        Returns a string representation of the TCP option.

        Returns:
        - str: Formatted string of the TCP option kind, length, and value.
        """
        return f'{self.kind}(length={self.length}, value={self.value})'

class TCPOption0Unit(TCPOptionUnit):
    """
    Represents the "End of Option List" TCP option (kind 0).

    Properties:
    - kind: Returns 'End of Option List'.
    - length: Returns the length attribute (empty by default for this option).
    - value: Returns the value attribute (empty by default for this option).
    """
    @property
    def kind(self) -> str:
        return 'End of Option List'
    @property
    def length(self) -> bytes:
        return self._length
    @property
    def value(self) -> bytes:
        return self._value

class TCPOption1Unit(TCPOptionUnit):
    """
    Represents the "No-Operation" TCP option (kind 1).

    Properties:
    - kind: Returns 'No-Operation'.
    - length: Returns the length attribute (empty by default for this option).
    - value: Returns the value attribute (empty by default for this option).
    """
    @property
    def kind(self) -> str:
        return 'No-Operation'
    @property
    def length(self) -> bytes:
        return self._length
    @property
    def value(self) -> bytes:
        return self._value

class TCPOption2Unit(TCPOptionUnit):
    """
    Represents the "Maximum Segment Size" TCP option (kind 2).

    Properties:
    - kind: Returns 'Maximum Segment Size'.
    - length: Interprets the length attribute as an integer.
    - value: Interprets the value as an integer representing the maximum segment size.
    """
    @property
    def kind(self) -> str:
        return 'Maximum Segment Size'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> int:
        return int.from_bytes(self._value, byteorder='big')

class TCPOption3Unit(TCPOptionUnit):
    """
    Represents the "Window Scale" TCP option (kind 3).

    Properties:
    - kind: Returns 'Window Scale'.
    - length: Interprets the length attribute as an integer.
    - value: Interprets the value as an integer representing the window scale factor.
    """
    @property
    def kind(self) -> str:
        return 'Window Scale'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> int:
        return int.from_bytes(self._value, byteorder='big')

class TCPOption4Unit(TCPOptionUnit):
    """
    Represents the "SACK Permitted" TCP option (kind 4).

    Properties:
    - kind: Returns 'SACK Permitted'.
    - length: Interprets the length attribute as an integer.
    - value: Returns the value attribute as bytes (typically unused for this option).
    """
    @property
    def kind(self) -> str:
        return 'SACK Permitted'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> bytes:
        return self._value

class TCPOption5Unit(TCPOptionUnit):
    """
    Represents the "SACK" (Selective Acknowledgment) TCP option (kind 5), which includes multiple
    acknowledgment ranges.

    Properties:
    - kind: Returns 'SACK'.
    - length: Interprets the length attribute as an integer.
    - value: Interprets the value as a list of TCPOptionSACKValue, representing SACK ranges.
    """
    @property
    def kind(self) -> str:
        return 'SACK'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> List[TCPOptionSACKValue]:
        sacks = []
        for i in range(0, len(self._value), 8):
            sack = {
                'left_edge': int.from_bytes(self._value[i:i+4], byteorder='big'),
                'right_edge': int.from_bytes(self._value[i+4:i+8], byteorder='big')
            }
            sacks.append(TCPOptionSACKValue(**sack))
        return sacks

class TCPOption8Unit(TCPOptionUnit):
    """
    Represents the "Timestamp" TCP option (kind 8), used for measuring round-trip time.

    Properties:
    - kind: Returns 'Timestamp'.
    - length: Interprets the length attribute as an integer.
    - value: Interprets the value as a dictionary containing timestamp and echo reply values.
    """
    @property
    def kind(self) -> str:
        return 'Timestamp'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> Dict[str, int]:
        values = {'timestamp': int.from_bytes(self._value[:4], byteorder='big'), 'echo_reply': int.from_bytes(self._value[4:], byteorder='big')}
        return values

class TCPUnknownOptionUnit(TCPOptionUnit):
    """
    Represents an unknown TCP option, allowing it to be stored and passed through without interpretation.

    Properties:
    - kind: Returns 'Unknown Option'.
    - length: Interprets the length attribute as an integer.
    - value: Returns the raw byte sequence of the unknown option.
    """
    @property
    def kind(self) -> str:
        return 'Unknown Option'
    @property
    def length(self) -> int:
        return int.from_bytes(self._length, byteorder='big')
    @property
    def value(self) -> bytes:
        return self._value


# -------------------TCP Option Unit Selector-------------------
class TCPOptionUnitSelectorHandler(SimpleCoRHandler):
    """
    Base handler class for selecting the appropriate TCP option unit class based on the option kind.
    Implements the **Chain of Responsibility** pattern, where each subclass handles a specific option kind.

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Attempts to handle the given option kind, returning the
      corresponding TCPOption unit class or passing the request to the next handler.

    Returns:
    - The matching TCPOption unit class if handled, otherwise `TCPUnknownOptionUnit`.
    """
    @abstractmethod
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if self._next_handler:
            return self._next_handler.handle(kind)
        return TCPUnknownOptionUnit

class TCPOption0UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption0Unit class for option kind 0 (End of Option List).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption0Unit` if the kind is 0; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 0:
            return TCPOption0Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption1UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption1Unit class for option kind 1 (No-Operation).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption1Unit` if the kind is 1; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 1:
            return TCPOption1Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption2UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption2Unit class for option kind 2 (Maximum Segment Size).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption2Unit` if the kind is 2; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 2:
            return TCPOption2Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption3UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption3Unit class for option kind 3 (Window Scale).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption3Unit` if the kind is 3; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 3:
            return TCPOption3Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption4UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption4Unit class for option kind 4 (SACK Permitted).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption4Unit` if the kind is 4; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 4:
            return TCPOption4Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption5UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption5Unit class for option kind 5 (SACK).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption5Unit` if the kind is 5; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 5:
            return TCPOption5Unit
        return super().handle(kind, *args, **kwargs)

class TCPOption8UnitSelectorHandler(TCPOptionUnitSelectorHandler):
    """
    Handler to select the TCPOption8Unit class for option kind 8 (Timestamp).

    Methods:
    - handle(kind: Any, *args, **kwargs) -> Any: Returns `TCPOption8Unit` if the kind is 8; otherwise,
      passes to the next handler.
    """
    def handle(self, kind: Any, *args, **kwargs) -> Any:
        if kind == 8:
            return TCPOption8Unit
        return super().handle(kind, *args, **kwargs)

class TCPOptionUnitSelector:
    """
    Utility class for selecting the appropriate TCP option unit class based on the option kind.
    Uses a chain of handlers to evaluate each option kind and return the corresponding TCPOption unit class.

    Methods:
    - select_tcp_option_unit(kind: int) -> type(TCPOptionUnit): Static method that initializes the handler
      chain and returns the appropriate TCPOption unit class based on the provided kind.
    """
    @staticmethod
    def select_tcp_option_unit(kind: int) -> type(TCPOptionUnit):
        """
        Selects the appropriate TCP option unit class based on the provided option kind.

        Parameters:
        - kind (int): The integer identifier for the TCP option kind.

        Returns:
        - type(TCPOptionUnit): The TCP option unit class corresponding to the specified kind, or `TCPUnknownOptionUnit`
          if the kind is unrecognized.
        """
        selectors = [
            TCPOption0UnitSelectorHandler(),
            TCPOption1UnitSelectorHandler(),
            TCPOption2UnitSelectorHandler(),
            TCPOption3UnitSelectorHandler(),
            TCPOption4UnitSelectorHandler(),
            TCPOption5UnitSelectorHandler(),
            TCPOption8UnitSelectorHandler(),
        ]
        selector_handler = selectors[0]
        for next_handler in selectors[1:]:
            selector_handler.set_next(next_handler)
            selector_handler = next_handler
        return selectors[0].handle(kind)

# -------------------TCP Option Header-------------------
class TCPOptionHeader(Header):
    """
    Represents the header of a TCP option and provides methods for constructing and retrieving
    the appropriate TCPOption unit based on the option kind.

    Methods:
    - _tcp_option_unit: Determines the correct TCPOption unit class using `TCPOptionUnitSelector`.
    - _construct: Constructs the option fields as a dictionary, packaging kind, length, and value as bytes.
    - unit: Returns an instance of the TCPOption unit, built from the constructed fields.

    Attributes:
    - _fields (dict): A dictionary holding 'kind', 'length', and 'value' for the TCP option.
    """
    def _tcp_option_unit(self) -> type(TCPOptionUnit):
        """
        Selects the appropriate TCPOption unit class based on the 'kind' field.

        Returns:
        - type(TCPOptionUnit): The TCPOption unit class determined by the option kind.
        """
        tcp_option_unit = TCPOptionUnitSelector.select_tcp_option_unit(self._fields['kind'])
        return tcp_option_unit

    def _construct(self) -> dict:
        """
        Constructs the fields of the TCP option header as a dictionary of packed bytes.

        Returns:
        - dict: Dictionary containing 'kind', 'length', and 'value' in byte format.
        """
        kind = struct.pack('!B', self._fields['kind'])
        length = struct.pack('!B', self._fields['length']) if self._fields.get('length') else b''
        value = self._fields['value'] if self._fields.get('value') else b''

        return {
            'kind': kind,
            'length': length,
            'value': value,
        }

    @property
    def unit(self) -> ProtocolUnit:
        """
        Returns an instance of the selected TCPOption unit, constructed from the header fields.

        Returns:
        - ProtocolUnit: The TCPOption unit instance initialized with the header's constructed fields.
        """
        tcp_option_unit = self._tcp_option_unit()
        return tcp_option_unit(**self._construct())


# -------------------TCP Option Header Builder-------------------
class TCPOptionHeaderBuilderInterface(HeaderBuilder, ABC):
    """
    Abstract builder interface for constructing TCP option headers, using the Builder pattern.

    Methods:
    - set_kind(kind: int): Abstract method to set the 'kind' field for a TCP option header.
    - set_length(length: int): Abstract method to set the 'length' field for a TCP option header.
    - set_value(value: int): Abstract method to set the 'value' field for a TCP option header.
    """
    @abstractmethod
    def set_kind(self, kind: int):
        pass
        # self._header.add_field('kind', kind)
    @abstractmethod
    def set_length(self, length: int):
        pass
        # self._header.add_field('length', length)
    @abstractmethod
    def set_value(self, value: int):
        pass
        # self._header.add_field('value', value)

class TCPOption0HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "End of Option List" TCP option (kind 0). This option does not have length or value fields.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 0.
    - set_length(length: int): Raises a ValueError if a length is provided, as this option has no length.
    - set_value(value: int): Raises a ValueError if a value is provided, as this option has no value.
    """
    def set_kind(self, kind: int = 0):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = None):
        if length is not None:
            raise ValueError('End of Option List option does not have a length')
        self._header.add_field('length', None)
    def set_value(self, value: int = None):
        if value is not None:
            raise ValueError('End of Option List option does not have a value')
        self._header.add_field('value', None)

class TCPOption1HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "No-Operation" TCP option (kind 1). This option does not have length or value fields.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 1.
    - set_length(length: int): Raises a ValueError if a length is provided, as this option has no length.
    - set_value(value: int): Raises a ValueError if a value is provided, as this option has no value.
    """
    def set_kind(self, kind: int = 1):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = None):
        if length is not None:
            raise ValueError('No-Operation option does not have a length')
        self._header.add_field('length', None)
    def set_value(self, value: int = None):
        if value is not None:
            raise ValueError('No-Operation option does not have a value')
        self._header.add_field('value', None)

class TCPOption2HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "Maximum Segment Size" TCP option (kind 2). This option has a default length of 4 bytes and a default
    value of 1460 (typical MSS value), but both can be customized.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 2 or uses a provided custom kind.
    - set_length(length: int): Sets the 'length' field to 4 by default or a custom length if provided.
    - set_value(value: int): Sets the 'value' field to 1460 (typical MSS) or a provided value in bytes.
    """
    def set_kind(self, kind: int = None):
        if kind is not None:
            self._header.add_field('kind', kind)
        else:
            self._header.add_field('kind', 2)
    def set_length(self, length: int = None):
        if length is not None:
            self._header.add_field('length', length)
        else:
            self._header.add_field('length', 4)
    def set_value(self, value: int = None):
        if value is not None:
            value_as_bytes = struct.pack('!H', value)
        else:
            value_as_bytes = struct.pack('!H', 1460)
        self._header.add_field('value', value_as_bytes)

class TCPOption3HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "Window Scale" TCP option (kind 3). This option has a default length of 3 bytes and a
    default value of 7 for the scaling factor, but both can be customized.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 3 by default.
    - set_length(length: int): Sets the 'length' field to 3 by default.
    - set_value(value: int): Sets the 'value' field to 7 by default, or a provided scaling factor in bytes.
    """
    def set_kind(self, kind: int = 3):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = 3):
        self._header.add_field('length', length)
    def set_value(self, value: int = 7):
        value_as_bytes = struct.pack('!B', value)
        self._header.add_field('value', value_as_bytes)

class TCPOption4HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "SACK Permitted" TCP option (kind 4). This option has a fixed length of 2 bytes
    and does not include a value.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 4 by default.
    - set_length(length: int): Sets the 'length' field to 2 by default.
    - set_value(value: int): Raises a ValueError if a value is provided, as this option does not have a value.
    """
    def set_kind(self, kind: int = 4):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = 2):
        self._header.add_field('length', length)
    def set_value(self, value: int = None):
        if value is not None:
            raise ValueError('SACK Permitted option does not have a value')
        self._header.add_field('value', None)

class TCPOption5HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "SACK" (Selective Acknowledgment) TCP option (kind 5). This option requires a length
    and includes multiple SACK ranges.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 5 by default.
    - set_length(length: int): Sets the 'length' field; raises an error if not provided.
    - set_value(value: List[TCPOptionSACKValue]): Constructs the value from provided SACK ranges.
    """
    def set_kind(self, kind: int = 5):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = None):
        """
        It's not necessary to set the length of the SACK option directly, as it is calculated based on the number of SACK ranges.
        """
        if length is None:
            raise ValueError('Length must be specified for SACK option')
        self._header.add_field('length', length)
    def set_value(self, value:List[TCPOptionSACKValue]):
        value_as_bytes = b''
        for sack in value:
            value_as_bytes += struct.pack('!I', sack.left_edge) + struct.pack('!I', sack.right_edge)
        self.set_length(len(value_as_bytes) + 2)
        self._header.add_field('value', value_as_bytes)

class TCPOption8HeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for the "Timestamp" TCP option (kind 8). This option has a default length of 10 bytes and
    includes a timestamp and an echo reply.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to 8 by default.
    - set_length(length: int): Sets the 'length' field to 10 by default.
    - set_value(value: TCPOptionTimestampValue): Constructs the value from a timestamp and echo reply.
    """
    def set_kind(self, kind: int = 8):
        self._header.add_field('kind', kind)
    def set_length(self, length: int = 10):
        self._header.add_field('length', length)
    def set_value(self, value: TCPOptionTimestampValue):
        value_as_bytes = struct.pack('!I', value.timestamp) + struct.pack('!I', value.echo_reply)
        self._header.add_field('value', value_as_bytes)

class TCPCustomOptionHeaderBuilder(TCPOptionHeaderBuilderInterface):
    """
    Builder for custom TCP options. Allows specifying arbitrary kinds, lengths, and values for options
    that do not fit standard option types.

    Methods:
    - set_kind(kind: int): Sets the 'kind' field to the specified value.
    - set_length(length: int): Sets the 'length' field; raises an error if not provided.
    - set_value(value: bytes): Sets the 'value' field as raw bytes and adjusts the length accordingly.

    Examples:
    - Maximum Segment Size (MSS) option (kind 2):
        ```
        value = 1460
        .set_value(value.to_bytes(2, byteorder='big'))
        Expected output: b'\x05\xB4'
        ```

    - Timestamp option (kind 8):
        ```
        value = 1328420247
        echo_reply = 3557171634
        .set_value(value.to_bytes(4, byteorder='big') + echo_reply.to_bytes(4, byteorder='big'))
        ```
    """
    def set_kind(self, kind: int):
        """
        Sets the 'kind' field for the custom option.

        Parameters:
        - kind (int): The kind identifier for the option.
        """
        self._header.add_field('kind', kind)
    def set_length(self, length: int = None):
        """
        Sets the 'length' field for the custom option. Raises an error if not provided.

        Parameters:
        - length (int): The length of the custom option.

        Raises:
        - ValueError: If length is not specified.
        """
        if length is None:
            raise ValueError('Length must be specified for custom option')
        self._header.add_field('length', length)
    def set_value(self, value: bytes):
        """
        Sets the 'value' field for the custom option as raw bytes, calculates and sets the length.

        Parameters:
        - value (bytes): The custom option value in bytes.

        Examples:
        - For an MSS option (kind 2), with a value of 1460:
            ```
            set_value(1460.to_bytes(2, byteorder='big'))
            ```
        - For a Timestamp option (kind 8), with timestamp and echo_reply values:
            ```
            timestamp = 1328420247
            echo_reply = 3557171634
            set_value(timestamp.to_bytes(4, byteorder='big') + echo_reply.to_bytes(4, byteorder='big'))
            ```
        """
        self.set_length(len(value) + 2)
        self._header.add_field('value', value)


# -------------------TCP Option Builder Director-------------------
class TCPOptionBuilderDirector:
    """
    Director class for constructing TCP option headers using a specified builder.
    This class orchestrates the construction process, configuring the option's kind, length, and value
    through a builder interface.

    Methods:
    - set_builder(builder: type(TCPOptionHeaderBuilderInterface)): Sets the builder to be used for constructing TCP options.
    - build(kind: int, length: int = None, value: Any = None): Configures and builds the TCP option header based on the provided parameters.

    Attributes:
    - _builder (TCPOptionHeaderBuilderInterface): The builder instance used for constructing the TCP option.
    """
    def __init__(self, builder: type(TCPOptionHeaderBuilderInterface)):
        """
        Initializes the director with a builder instance for constructing a TCP option header.

        Parameters:
        - builder (type(TCPOptionHeaderBuilderInterface)): The builder class to use for constructing the option.
        """
        self._builder = builder(TCPOptionHeader())

    def set_builder(self, builder: type(TCPOptionHeaderBuilderInterface)):
        """
        Sets a new builder for constructing the TCP option header.

        Parameters:
        - builder (type(TCPOptionHeaderBuilderInterface)): The new builder class to use.
        """
        self._builder = builder(TCPOptionHeader())

    def build(self, kind: int, length: int = None, value: Any = None):
        """
        Constructs the TCP option header by setting the kind, length, and value using the builder.

        Parameters:
        - kind (int): The kind identifier for the TCP option.
        - length (int, optional): The length of the option. If None, uses the default length defined by the builder.
        - value (Any, optional): The value of the option. If None, uses the default value defined by the builder.

        Returns:
        - TCPOptionHeader: The constructed TCP option header instance.
        """
        self._builder.set_kind(kind)
        self._builder.set_length(length)
        self._builder.set_value(value)
        return self._builder.build()



from __future__ import annotations
from abc import ABC

from ttlinks.common.design_template.cor import TrackedCoRHandler


class IPv4ParserHandler(TrackedCoRHandler, ABC):
    """
    Abstract base class for all IPv4 field parser handlers.
    Inherits from TrackedCoRHandler, which presumably manages the Chain of Responsibility (CoR)
    and provides tracking through the `_tracker` attribute.
    """
    pass

class IPv4VersionAndIHLFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Version and Internet Header Length (IHL) field from an IPv4 packet.
    The first byte of the packet contains both the version (first 4 bits) and IHL (last 4 bits).
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['version_and_ihl'] = packet[0:1]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)


class IPv4TOSFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Type of Service (TOS) field from an IPv4 packet.
    The second byte (index 1) contains the TOS, which includes DSCP and ECN fields.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['tos'] = packet[1:2]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4TotalLengthFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Total Length field from an IPv4 packet.
    The total length is a 2-byte field located at bytes 2 and 3.
    It represents the entire length of the IPv4 packet (header + payload).
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['total_length'] = packet[2:4]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4IdentificationFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Identification field from an IPv4 packet.
    The identification is a 2-byte field located at bytes 4 and 5.
    It is used to identify fragments of an original packet.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['identification'] = packet[4:6]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4FlagsAndFragmentOffsetFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Flags and Fragment Offset field from an IPv4 packet.
    This is a 2-byte field located at bytes 6 and 7. The first 3 bits are the flags,
    and the remaining 13 bits are the fragment offset.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['flags_and_fragment_offset'] = packet[6:8]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4TTLFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Time to Live (TTL) field from an IPv4 packet.
    The TTL field is a 1-byte field located at byte 8. It determines the maximum
    number of hops the packet can take before being discarded.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['ttl'] = packet[8:9]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4ProtocolFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Protocol field from an IPv4 packet.
    The Protocol field is a 1-byte field located at byte 9. It indicates the protocol
    of the encapsulated payload (e.g., ICMP, TCP, UDP).
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['protocol'] = packet[9:10]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4HeaderChecksumFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Header Checksum field from an IPv4 packet.
    The Header Checksum is a 2-byte field located at bytes 10 and 11. It is used to
    verify the integrity of the IPv4 header.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['header_checksum'] = packet[10:12]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4SourceAddressFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Source Address field from an IPv4 packet.
    The Source Address is a 4-byte field located at bytes 12 to 15. It contains
    the IPv4 address of the packet's sender.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['source_address'] = packet[12:16]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4DestinationAddressFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Destination Address field from an IPv4 packet.
    The Destination Address is a 4-byte field located at bytes 16 to 19. It contains
    the IPv4 address of the packet's intended recipient.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            self._tracker['destination_address'] = packet[16:20]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4OptionsFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Options field from an IPv4 packet.
    The Options field is variable-length and starts after the fixed 20-byte header.
    The length of the options field is determined by the Internet Header Length (IHL),
    which specifies the total header length in 4-byte words. If IHL > 5, options are present.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            ihl = self._tracker.get('ihl', 5) * 4
            self._tracker['options'] = packet[20:ihl]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4PayloadFieldParserHandler(IPv4ParserHandler):
    """
    Handler for parsing the Payload (data) field from an IPv4 packet.
    The payload begins immediately after the IPv4 header and options. The header length
    is determined by the Internet Header Length (IHL) field, and the payload starts from that offset.
    """
    def handle(self, packet: bytes, *args, **kwargs):
        if isinstance(packet, bytes):
            ihl = self._tracker.get('ihl', 5) * 4
            self._tracker['payload'] = packet[ihl:]
        if self._next_handler:
            return self._next_handler.handle(packet, *args, **kwargs)

class IPv4PacketParser:
    """
    Singleton parser class for parsing an IPv4 packet field by field using the Chain of Responsibility (CoR) pattern.
    The class ensures that only one instance of the parser exists at a time.
    """
    __instance: IPv4PacketParser = None

    def __new__(cls):
        """
        Implements the Singleton pattern. Ensures that only one instance of IPv4PacketParser is created.
        If an instance doesn't exist, a new one is created. If it exists, the existing instance is returned.
        """
        if cls.__instance is None:
            cls.__instance = super(IPv4PacketParser, cls).__new__(cls)
        return cls.__instance

    def _reset_parser(self):
        """
        Resets the parser chain by initializing all the field parser handlers in sequence.
        Each handler is responsible for parsing a specific field from the IPv4 packet.
        The handlers are linked together using the set_next() method to form a chain of responsibility.
        """
        self._packet_parser = IPv4VersionAndIHLFieldParserHandler()
        (
            self._packet_parser
            .set_next(IPv4TOSFieldParserHandler())
            .set_next(IPv4TotalLengthFieldParserHandler())
            .set_next(IPv4IdentificationFieldParserHandler())
            .set_next(IPv4FlagsAndFragmentOffsetFieldParserHandler())
            .set_next(IPv4TTLFieldParserHandler())
            .set_next(IPv4ProtocolFieldParserHandler())
            .set_next(IPv4HeaderChecksumFieldParserHandler())
            .set_next(IPv4SourceAddressFieldParserHandler())
            .set_next(IPv4DestinationAddressFieldParserHandler())
            .set_next(IPv4OptionsFieldParserHandler())
            .set_next(IPv4PayloadFieldParserHandler())
        )

    def parse(self, frame: bytes):
        """
        Parses the given IPv4 packet (frame) by passing it through the chain of handlers.

        Args:
            frame (bytes): The raw IPv4 packet data in bytes.

        Returns:
            dict: A dictionary of parsed IPv4 fields stored in the handler's tracker.
        """
        self._reset_parser()
        self._packet_parser.handle(frame)
        return self._packet_parser.get_tracker()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any


class CoRHandler(ABC):
    """Abstract base class for handlers in the Chain of Responsibility pattern."""

    @abstractmethod
    def set_next(self, h: CoRHandler) -> CoRHandler:
        """Set the next handler in the chain."""
        pass

    @abstractmethod
    def handle(self, request: Any, *args, **kwargs) -> Any:
        """Handle the request. Concrete handlers must implement this method."""
        pass


class SimpleCoRHandler(CoRHandler, ABC):
    def __init__(self):
        self._next_handler = None  # Reference to the next handler in the chain

    def set_next(self, h: SimpleCoRHandler) -> SimpleCoRHandler:
        """
        Sets the next handler in the chain.

        Args:
            h (SimpleCoRHandler): The next handler to link.

        Returns:
            SimpleCoRHandler: The next handler (for chaining).

        Raises:
            TypeError: If the provided handler is not an instance of SimpleCoRHandler.
        """
        if not isinstance(h, SimpleCoRHandler):
            raise TypeError("The next handler must be an instance of SimpleCoRHandler.")
        self._next_handler = h
        return h

    def get_next(self) -> SimpleCoRHandler:
        """Returns the next handler in the chain."""
        return self._next_handler

    @abstractmethod
    def handle(self, request: Any, *args, **kwargs):
        """
        Handle the request or pass it to the next handler in the chain.
        Subclasses must implement this method.
        """
        pass


class BidirectionalCoRHandler(CoRHandler, ABC):
    _next_handler = None  # Reference to the next handler in the chain
    _previous_handler = None  # Reference to the previous handler in the chain

    def set_next(self, h: BidirectionalCoRHandler) -> BidirectionalCoRHandler:
        """
        Sets the next handler in the chain and links the previous handler.

        Args:
            h (BidirectionalCoRHandler): The next handler to link.

        Returns:
            BidirectionalCoRHandler: The next handler (for chaining).

        Raises:
            TypeError: If the provided handler is not an instance of BidirectionalCoRHandler.
        """
        if not isinstance(h, BidirectionalCoRHandler):
            raise TypeError("The next handler must be an instance of BidirectionalCoRHandler.")
        self._next_handler = h
        h._previous_handler = self
        return h

    def get_next(self) -> BidirectionalCoRHandler:
        """Returns the next handler in the chain."""
        return self._next_handler

    def get_previous(self) -> BidirectionalCoRHandler:
        """Returns the previous handler in the chain."""
        return self._previous_handler

    @abstractmethod
    def handle(self, request, *args, **kwargs):
        """
        Handle the request or pass it to the next handler in the chain.
        Subclasses must implement this method.
        """
        pass

class TrackedCoRHandler(CoRHandler, ABC):
    def __init__(self):
        self._next_handler = None  # Reference to the next handler in the chain
        self._tracker = {}  # Dictionary to track the request and response

    def _set_tracker(self, tracker: dict):
        self._tracker = tracker

    def set_next(self, h: TrackedCoRHandler) -> TrackedCoRHandler:
        if not isinstance(h, TrackedCoRHandler):
            raise TypeError("The next handler must be an instance of TrackedCoRHandler.")
        self._next_handler = h
        self._next_handler._set_tracker(self._tracker)
        return h

    def get_next(self) -> TrackedCoRHandler:
        return self._next_handler

    def get_tracker(self) -> dict:
        return self._tracker

    @abstractmethod
    def handle(self, request: Any, *args, **kwargs) -> Any:
        pass

class ListBasedCoRHandler(CoRHandler, ABC):
    def __init__(self):
        self._next_handler = None  # Reference to the next handler in the chain
        self._items = []  # List to track the request and response

    def _set_items(self, items: list):
        self._items = items

    def set_next(self, h: ListBasedCoRHandler) -> ListBasedCoRHandler:
        if not isinstance(h, ListBasedCoRHandler):
            raise TypeError("The next handler must be an instance of ListBasedCoRHandler.")
        self._next_handler = h
        self._next_handler._set_items(self._items)
        return h

    def get_next(self) -> ListBasedCoRHandler:
        return self._next_handler

    def get_items(self) -> list:
        return self._items

    @abstractmethod
    def handle(self, request: Any, *args, **kwargs) -> Any:
        pass

class ProtocolUnitSelectorCoRHandler(CoRHandler, ABC):
    def __init__(self, parser: Any = None):
        self._next_handler = None  # Reference to the next handler in the chain
        self._parser = parser  # Reference to the parser object
        self._parsed_data = None  # None if not parsed, dict if parsed

    def _set_parser(self, parser: Any):
        """Sets the parser object."""
        self._parser = parser

    def set_parsed_data(self, parsed_data: dict):
        """Sets the parsed flag."""
        self._parsed_data = parsed_data

    def set_next(self, h: ProtocolUnitSelectorCoRHandler) -> ProtocolUnitSelectorCoRHandler:
        if not isinstance(h, ProtocolUnitSelectorCoRHandler):
            raise TypeError("The next handler must be an instance of ProtocolParserCoRHandler.")
        self._next_handler = h
        self._next_handler._set_parser(self._parser)
        # self._next_handler._set_parsed_data(self._parsed_data)
        return h

    def get_next(self) -> ProtocolUnitSelectorCoRHandler:
        return self._next_handler

    def get_parser(self) -> Any:
        return self._parser

    @abstractmethod
    def handle(self, request: bytes, *args, **kwargs) -> Any:
        pass


from __future__ import annotations

import re


class NumeralConverter:

    @staticmethod
    def binary_to_decimal(binary_string: str) -> int:
        """
        Converts a binary string to its decimal (base-10) equivalent.

        Parameters:
        binary_string (str): A binary string to be converted.

        Returns:
        int: The decimal equivalent of the binary string.

        Raises:
        TypeError: If the input is not a string.
        """
        if type(binary_string) is not str:
            raise TypeError("Binary must be a string.")
        return int(binary_string, 2)

    @staticmethod
    def decimal_to_binary(decimal: int, r_just: int = 8) -> str:
        """
        Converts a decimal (base-10) number to its binary string equivalent,
        with optional right-justified padding.

        Parameters:
        decimal (int): The decimal number to convert.
        r_just (int): The number of digits to right-justify the binary output (default is 8).

        Returns:
        str: The binary string equivalent of the decimal number.

        Raises:
        TypeError: If the input types are not as expected (int for both parameters).
        """
        if type(decimal) is not int:
            raise TypeError("decimal must be an int.")
        if type(r_just) is not int:
            raise TypeError("r_just must be an int.")
        return bin(decimal)[2:].rjust(r_just, '0')

    @staticmethod
    def binary_to_hexadecimal(binary_string: str) -> str:
        """
        Converts a binary string to its hexadecimal string equivalent.

        Parameters:
        binary (str): A binary string to be converted.

        Returns:
        str: The hexadecimal string equivalent of the binary number.

        Raises:
        TypeError: If the input is not a string.
        """
        if type(binary_string) is not str:
            raise TypeError("binary must be a string.")
        decimal_value = int(binary_string, 2)
        hexadecimal_string = hex(decimal_value)[2:]
        return hexadecimal_string.upper()

    @staticmethod
    def hexadecimal_to_binary(hexadecimal: str, r_just: int = 8) -> str:
        """
        Converts a hexadecimal string to its binary string equivalent, with optional right-justified padding.

        Parameters:
        hexadecimal (str): A hexadecimal string to be converted.
        r_just (int): The number of digits to right-justify the binary output (default is 8).

        Returns:
        str: The binary string equivalent of the hexadecimal number.

        Raises:
        TypeError: If the input types are not as expected (string for hexadecimal, int for r_just).
        """
        if type(hexadecimal) is not str:
            raise TypeError("hexadecimal is not a string.")
        if not re.match(r"^[0-9A-F]+$", hexadecimal.upper()):
            raise TypeError("wrong hexadecimal format.")
        if type(r_just) is not int:
            raise TypeError("r_just is not a int.")
        decimal_value = int(hexadecimal, 16)
        binary_string = bin(decimal_value)[2:].rjust(r_just, '0')
        return binary_string

    @staticmethod
    def hexadecimal_to_decimal(hexadecimal: str) -> int:
        """
        Converts a hexadecimal string to its decimal (base-10) equivalent.

        Parameters:
        hexadecimal (str): A hexadecimal string to be converted.

        Returns:
        int: The decimal equivalent of the hexadecimal number.

        Raises:
        TypeError: If the input is not a string.
        """
        if type(hexadecimal) is not str:
            raise TypeError("hexadecimal must be a string.")
        if not re.match(r"^[0-9A-F]+$", hexadecimal.upper()):
            raise TypeError("wrong hexadecimal format.")
        return int(hexadecimal, 16)

    @staticmethod
    def decimal_to_hexadecimal(decimal: int, r_just: int = 2) -> str:
        """
        Converts a decimal (base-10) number to its hexadecimal string equivalent,
        with optional right-justified padding.

        Parameters:
        decimal (int): The decimal number to convert.
        r_just (int): The number of digits to right-justify the hexadecimal output (default is 2).

        Returns:
        str: The hexadecimal string equivalent of the decimal number.

        Raises:
        TypeError: If the input types are not as expected (int for both parameters).
        """
        if type(decimal) is not int:
            raise TypeError("decimal must be an int.")
        return hex(decimal)[2:].upper().rjust(r_just, '0')

    @staticmethod
    def bytes_to_decimal(byte_string: bytes) -> int:
        """
        Converts a bytes object to its decimal (base-10) equivalent.

        Parameters:
        byte_string (bytes): A bytes object to be converted.

        Returns:
        int: The decimal equivalent of the bytes object.

        Raises:
        TypeError: If the input is not a bytes object.
        """
        if type(byte_string) is not bytes:
            raise TypeError("byte_string must be a bytes object.")
        return int.from_bytes(byte_string, byteorder='big')